/* -*-c-*- */

/*
  lex input file for source code formatting of C.

  Adapted to libformat by Dave Bailey <dave@daveb.net>

  Rewritten by Robert Helgesson <rycee@home.se>
*/

%{
#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <stdio.h>
#include "markup.h"
#include "syntax.h"

#define YY_NO_UNPUT 1

/* these macros will end up saving us a lot of space. */

#define CAT     sh_cats(&gData,format_c_yytext)
#define XPRE(a) sh_cats(&gData,"<font class=\"%s\">",(a))
#define PRE(a)  XPRE(a); CAT
#define XCAT(a) sh_cats(&gData,(a))
#define POST    XCAT("</font>")
#define WRAP(a) PRE(a); POST

#define CATC      sh_catc(&gData,format_c_yytext[0]); format_c_yytext++;
#define IWRAP(a)  CATC; WRAP(a)

/* simple helper */
#define DP(a)     PRE(a); BEGIN(preproc)

static sh_data gData;

char *c_format ( const char *str )
{
  YY_BUFFER_STATE  yybs   = NULL;
  char            *output = NULL;

  format_c_yy_scan_string(str);
  XCAT(START_TEXT);
  format_c_yylex();
  XCAT(END_TEXT);

  if ( gData.buf ) output = strdup(gData.buf);

  yy_delete_buffer(yybs);
  sh_reset(&gData);
  return output;
}

%}

%option stack
%option noyywrap
%option prefix="format_c_yy"

WS [ \t]
BOUND [^a-zA-Z0-9_"'#]

BASIC_TYPE char|short|int|long|signed|unsigned|float|double|void
KEYWORD typedef|sizeof|case|default|auto|register|static|extern|const|volatile
KEYWORD_FLOW if|else|switch|while|do|for|goto|return|break|continue
TYPE_SPEC enum|struct|union

PREPROC #line|#pragma|#error|#warning
PREPROC_STARTCOND #if|#ifdef|#ifndef
PREPROC_COND #else|#elif
PREPROC_ENDCOND #endif
PREPROC_INCLUDE #include
PREPROC_DEFINE #define|#undef

%s preproc
%s ccomment
%s quote
%s squote

%%

    /********************
     * General
     ********************/

("http://"|"mailto:"|"news://"|"ftp://"|"gopher://")[^<>\"\'%|\\^\{\}` \n\t]* {
  XCAT("<a href=\""); CAT; XCAT("\">"); CAT; XCAT("</a>");
}

    /********************
     * Comments
     ********************/

<INITIAL>"/*"   |
<preproc>"/*"   { PRE(COMMENT); yy_push_state(ccomment); }
<ccomment>"*/"  { CAT; POST; yy_pop_state();             }

    /********************
     * Preprocessor
     ********************/

<INITIAL>^{WS}*{PREPROC}                            { DP(PREPROC);            }
<INITIAL>^{WS}*{PREPROC_STARTCOND}                  { DP(PREPROC_STARTCOND);  }
<INITIAL>^{WS}*{PREPROC_COND}                       { DP(PREPROC_COND);       }
<INITIAL>^{WS}*{PREPROC_ENDCOND}                    { DP(PREPROC_ENDCOND);    }
<INITIAL>^{WS}*{PREPROC_DEFINE}                     { DP(PREPROC_DEFINE);     }
<INITIAL>^{WS}*{PREPROC_INCLUDE}({WS}+\"[^\"]+\")?  { DP(PREPROC);            }

<preproc>{
  \\{WS}*\n     { CAT; }
  .?$           { CATC; POST; BEGIN(INITIAL); }
}

    /********************
     * Constants
     ********************/

<INITIAL>"\""   |
<preproc>"\""   { PRE(STRING_LITERAL); yy_push_state(quote); }
<quote>{
  "\\\""        { CAT; }
  "\\\\"        { CAT; }
  "\""          { CATC; POST; yy_pop_state(); }
}

<INITIAL>"\'"   |
<preproc>"\'"   { PRE(CHAR_LITERAL); yy_push_state(squote);  }
<squote>{
  "\\\'"        { CAT; }
  "\\\\"        { CAT; }
  "\'"          { CATC; POST; yy_pop_state(); }
}

    /********************
     * C syntax
     ********************/

<INITIAL>{BOUND}{BASIC_TYPE}/{BOUND}      { IWRAP(BASIC_TYPE);      }
<INITIAL>{BOUND}{KEYWORD_FLOW}/{BOUND}    { IWRAP(KEYWORD_FLOW);    }
<INITIAL>{BOUND}{KEYWORD}/{BOUND}         { IWRAP(KEYWORD);         }
<INITIAL>{BOUND}{TYPE_SPEC}/{BOUND}       { IWRAP(TYPE_SPECIFIER);  }

    /********************
     * Whatever is left
     ********************/

\n              { CATC; }
.               { CATC; }
