#include <config.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include "format.h"

/* function bodies are in x2html.l */
extern char *c_format(const char *);
extern char *cpp_format(const char *);
extern char *java_format(const char *);
extern char *python_format(const char *);
extern char *verilog_format(const char *);
extern char *vhdl_format(const char *);

static struct {
  char *(*func)(const char *);
  const char *language;
} formatters[] = {
  { c_format,       "c" },
  { cpp_format,     "c++" },
  { java_format,    "java" },
  { python_format,  "python" },
  { verilog_format, "verilog" },
  { vhdl_format,    "vhdl" },
};

static int  languages = sizeof(formatters)/sizeof(formatters[0]);

/* static helper functions */

static char *format ( const char *input, const char *language );
static char *gobble ( int fd );

/* the real meat */

int
format_valid_lang(const char *language)
{
  int i;

  for ( i = 0; i < languages; i++ ) {
    if ( strcmp(formatters[i].language, language) == 0 ) {
      return 1;
    }
  }

  return 0;
}

char *
format_fd ( int fd, const char *language )
{
  char *input;
  char *output = NULL;

  if ( fd < 0 ) return NULL;

  if ( (input = gobble(fd)) ) {
    output = format(input, language);
    free(input);
  }

  return output;
}

char *
format_fp ( FILE *fp, const char *language )
{
  int fd;

  if ( ! fp ) return NULL;

  if ( (fd = fileno(fp)) == -1 ) {
    fprintf(stderr, "format_fp: %s\n", strerror(errno));
    return NULL;
  }

  return format_fd(fd, language);
}

char *
format_file ( const char *filename, const char *language )
{
  char *input;
  char *output = NULL;
  int fd;

  if ( ! filename ) return NULL;

  if ( (fd = open(filename, O_RDONLY)) == -1 ) {
    fprintf(stderr, "open[%s]: %s\n", filename, strerror(errno));
    return NULL;
  }

  if ( (input = gobble(fd)) ) {
    output = format(input,language);
    free(input);
  }

  close(fd);

  return output;
}

char *
format_string ( const char *str, const char *language )
{
  return format(str,language);
}


/* this function gets called by both the format functions */

static char *
format ( const char *input, const char *language )
{
  char *output = NULL;
  int i;

  if ( !input || !language ) return NULL;

  for ( i = 0; i < languages; i++ ) {
    if ( strcmp(formatters[i].language, language) == 0 ) {
      output = (formatters[i].func)(input);
      break;
    }
  }

  return output;
}


static char *
gobble ( int fd )
{
  char *contents = NULL;
  struct stat sb;
  ssize_t br;

  if ( fstat(fd, &sb) == -1 ) {
    fprintf(stderr, "gobble: stat: %s\n", strerror(errno));
    return NULL;
  }

  if ( (contents = calloc(sb.st_size + 1, 1)) == NULL ) {
    fprintf(stderr, "gobble: calloc: %s\n", strerror(errno));
    return NULL;
  }

  if ( (br = read(fd, contents, sb.st_size)) == -1 ) {
    free(contents);
    fprintf(stderr, "gobble: read[%d]: %s", (int) sb.st_size, strerror(errno));
    return NULL;
  }

  return contents;
}
