/* -*-c-*- */

/*
  lex input file for source code formatting of Java.

  Adapted to libformat by Dave Bailey <dave@daveb.net>
*/

%{
#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <stdio.h>
#include "markup.h"
#include "syntax.h"
#include "format.h"

#define JAVA_LANG "java_lang"

#define YY_NO_UNPUT 1

/* these macros will end up saving us a lot of space. */

#define CAT     sh_cats(&gData,format_java_yytext)
#define XPRE(a) sh_cats(&gData,"<font class=\"%s\">",(a))
#define PRE(a)  XPRE(a); CAT
#define XCAT(a) sh_cats(&gData,(a))
#define POST    XCAT("</font>")
#define WRAP(a) PRE(a); POST

#define CATC      sh_catc(&gData,format_java_yytext[0]); format_java_yytext++;
#define IWRAP(a)  CATC; WRAP(a)


static sh_data gData;

char *java_format ( const char *str )
{
  YY_BUFFER_STATE  yybs   = NULL;
  char            *output = NULL;

  format_java_yy_scan_string(str);
  XCAT(START_TEXT);
  format_java_yylex();
  XCAT(END_TEXT);

  if ( gData.buf ) output = strdup(gData.buf);

  yy_delete_buffer(yybs);
  sh_reset(&gData);
  return output;
}

%}

%option stack
%option noyywrap
%option prefix="format_java_yy"

WS [ \t]
BOUND [^a-zA-Z0-9_"']

BASIC_TYPE boolean|char|byte|short|int|long|float|double|void
JAVA_LANG Cloneable|Runnable|Boolean|Byte|Character|Class|ClassLoader|Compiler|Double|Float|Integer|Long|Math|Number|Object|Process|Runtime|SecurityManager|Short|String|StringBuffer|System|Thread|ThreadGroup|Throwable|Void|ArithmeticException|ArrayIndexOutOfBoundsException|ArrayStoreException|ClassCastException|ClasNotFoundException|CloneNotSupportedException|Exception|IllegalAccessException|IlegalArgumentException|IllegalMonitorStateException|IllegalStateException|IllegalThreadStateException|IndexOutOfBoundsException|InstantiationException|InterruptedException|NegativeArraySizeException|NoSuchFieldException|NoSuchMethodException|NullPointerException|NumberFormatException|RuntimeException|SecurityException|StringIndexOutOfBoundsException|AbtractMethodError|ClassCircularityError|ClassFormatError|Error|ExceptionInInitializerError|IllegalAccessError|IncompatibleClassChangeError|InstantiationError|InternalError|LinkageError|NoClassDefFoundError|NoSuchFieldError|NoSuchMethodError|OutOfMemoryError|StackOverflowError|ThreadDeath|UnknownError|UnsatisfiedLinkError|VerifyError|VirtualMachineError
KEYWORD instanceof|abstract|final|native|static|synchronized|transient|volatile|default|operator|outer|this|implements|package|throws|byvalue|extends|inner|case|true|false|cast|rest|var|native|const|future|new|super|generic|null
KEYWORD_FLOW if|else|switch|while|do|for|try|catch|finally|goto|break|return|continue
ACCESS_SPEC public|protected|private
TYPE_SPEC class|interface
PREPROC_INCLUDE import

%s ccomment
%s quote
%s squote

%%

    /********************
     * General
     ********************/

("http://"|"mailto:"|"news://"|"ftp://"|"gopher://")[^<>\"\'%|\\^\{\}` \n\t]* {
  XCAT("<a href=\""); CAT; XCAT("\">"); CAT; XCAT("</a>");
}

    /********************
     * Comments
     ********************/

<INITIAL>"//".*   { WRAP(COMMENT); }

<INITIAL>"/*"   { PRE(COMMENT); yy_push_state(ccomment); }
<ccomment>"*/"  { CAT; POST; yy_pop_state(); }

    /********************
     * Constants
     ********************/

<INITIAL>"\""   { PRE(STRING_LITERAL); yy_push_state(quote); }
<quote>{
  "\\\""        { CAT; }
  "\\\\"        { CAT; }
  "\""          { CATC; POST; yy_pop_state(); }
}

<INITIAL>"'"    { PRE(CHAR_LITERAL); yy_push_state(squote);  }
<squote>{
  "\\'"         { CAT; }
  "\\\\"        { CAT; }
  "\'"          { CATC; POST; yy_pop_state(); }
}

    /********************
     * Java syntax
     ********************/

<INITIAL>{BOUND}{BASIC_TYPE}/{BOUND}       { IWRAP(BASIC_TYPE);       }
<INITIAL>{BOUND}{JAVA_LANG}/{BOUND}        { IWRAP(JAVA_LANG);        }
<INITIAL>{BOUND}{KEYWORD}/{BOUND}          { IWRAP(KEYWORD);          }
<INITIAL>{BOUND}{ACCESS_SPEC}/{BOUND}      { IWRAP(ACCESS_SPECIFIER); }
<INITIAL>{BOUND}{TYPE_SPEC}/{BOUND}        { IWRAP(TYPE_SPECIFIER);   }
<INITIAL>{BOUND}{KEYWORD_FLOW}/{BOUND}     { IWRAP(KEYWORD_FLOW);     }
<INITIAL>^{WS}*{PREPROC_INCLUDE}.*         { WRAP(PREPROC);           }

    /********************
     * Whatever is left
     ********************/

\n             { CATC; }
.              { CATC; }
