#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include "format.h"
#include "config.h"

/* #defines */

#define STYLE_SHEET \
"/* Shared among languages */              \n" \
"                                          \n" \
".format_global         { background-color: #eee; color: #000 }\n" \
"                                          \n" \
"FONT.basic_type        { color: #604020 } \n" \
"FONT.keyword_flow      { color: #e08000 } \n" \
"FONT.keyword           { color: #008000 } \n" \
"FONT.access_specifier  { color: #ff0000 } \n" \
"FONT.type_specifier    { color: #0000ff } \n" \
"FONT.comment           { color: #800000 } \n" \
"FONT.preproc           { color: #806020 } \n" \
"FONT.preproc_startcond { color: #228b22 } \n" \
"FONT.preproc_cond      { color: #806020 } \n" \
"FONT.preproc_endcond   { color: #228b22 } \n" \
"FONT.preproc_define    { color: #228b22 } \n" \
"FONT.string_literal    { color: #002080 } \n" \
"FONT.char_literal      { color: #008080 } \n" \
"                                          \n" \
"/* Java */                                \n" \
"                                          \n" \
"FONT.java_lang         { color: #8b008b } \n" \
"                                          \n" \
"/* Python */                              \n" \
"                                          \n" \
"FONT.python_exception  { color: #8b008b } \n" \
"                                          \n" \
"/* Verilog */                             \n" \
"                                          \n" \
"FONT.verilog_kw        { color: #008000 } \n" \
"FONT.verilog_bt        { color: #8b008b } \n" \
"                                          \n" \
"/* VHDL */                                \n" \
"                                          \n" \
"FONT.vhdl_function     { color: #8b008b } \n" \
"FONT.vhdl_package      { color: #ff0000 } \n" \
"FONT.vhdl_attribute    { color: #b22222 } \n" \
"FONT.vhdl_enum         { color: #a020f0 } \n" \
"FONT.vhdl_constant     { color: #902080 } \n"


/* typedefs */

typedef struct fmt_lang fmt_lang;

struct fmt_lang {
  char lang[32];
  char name[32];
  char *ext[8];
};

/* externs */

extern char *optarg;
extern int   optind;

/* static variables */

static fmt_lang gLang[] = {
  { "c",       "C",       { "c", NULL } },
  { "c++",     "C++",     { "h","cc","cpp","hpp","cxx","hxx","C", NULL } },
  { "java",    "Java",    { "java", NULL } },
  { "python",  "Python",  { "py", NULL } },
  { "verilog", "Verilog", { "v", NULL } },
  { "vhdl",    "VHDL",    { "vhd", NULL } }
};

static int gNlang = sizeof(gLang)/sizeof(fmt_lang);

/* static functions */

static fmt_lang *get_lang        ( char *lang );
static fmt_lang *get_lang_by_ext ( char *ext );
static void      print_langs     ( void );

/* main program */

int
main ( int argc, char **argv )
{
  char       ch;
  char     * output;
  char     * progname;
  char     * ext;
  char       language[256];
  char       outfile[BUFSIZ];
  int        have_lang = 0;
  fmt_lang * this_lang = NULL;
  FILE     * fp;
  int        i;

  progname = argv[0];
  memset(language,0,sizeof(language));

  while ( (ch = getopt(argc,argv,"l:")) != -1 ) {
    switch (ch) {
    case 'l':
      strncpy(language,optarg,sizeof(language));
      break;
    }
  }
  argc -= optind;
  argv += optind;

  if ( argc < 1 ) {
    printf("usage: %s [-l language] <filename> ...\n",
	   progname);
    print_langs();
    exit(1);
  }

  if ( (have_lang = language[0]) ) {
    
    /* fix c++ if necessary */

    if ( !strcmp(language,"cpp") ) 
      strncpy(language,"c++",sizeof(language));

    this_lang = get_lang(language);
  }

  if ( have_lang && !this_lang ) {
    printf("%s: unrecognized language %s\n",progname,language);
    print_langs();
    exit(2);
  }

  /* loop through input files */

  for ( i = 0; i < argc; i++ ) {

    if ( !have_lang ) {
      if ( (ext = strrchr(argv[i],'.')) && *(ext+1) ) {
	if ( !(this_lang = get_lang_by_ext(ext+1)) ) {
	  printf("%s: unsupported extension [%s] "
		 "for language autodetection\n", ext, argv[i]);
	}
      } else {
	printf("%s: unable to find a suffix "
	       "for language autodetection\n", argv[i]);
      }
    }

    if ( this_lang ) {
      output = format_file(argv[i],this_lang->lang);
      if ( output ) {
	snprintf(outfile,sizeof(outfile),"%s.html",argv[i]);
	printf("Syntax highlighted %s as %s, saving to %s\n",
	       argv[i],this_lang->name,outfile);

	if ( (fp = fopen(outfile,"w")) ) {
	  fprintf(fp,"<html>\n");
	  fprintf(fp,"<head>\n");
	  fprintf(fp,"<style type=\"text/css\"><!--\n");
	  fprintf(fp,STYLE_SHEET);
	  fprintf(fp,"--></style>\n");
	  fprintf(fp,"</head>\n");
	  fprintf(fp,"<body class=\"format_global\">\n");
	  fprintf(fp,"%s",output);
	  fprintf(fp,"\n</body>\n</html>\n");
	  fclose(fp);
	} else {
	  printf("Saving to %s failed: %s\n",outfile,strerror(errno));
	}
	free(output);
      } else {
	printf("format_file(%s,%s) failed.\n",argv[i],this_lang->lang);
      }
    }
  }

  return 0;
}


static fmt_lang *
get_lang ( char *lang ) 
{
  int         i;
  fmt_lang  * this_lang = NULL;
  
  for ( i = 0; i < gNlang; i++ ) {
    if ( !strcmp(lang,gLang[i].lang) ) {
      this_lang = &gLang[i];
      break;
    }
  }

  return this_lang;
}



static fmt_lang *
get_lang_by_ext ( char *ext ) 
{
  int         i;
  int         j;
  fmt_lang  * this_lang = NULL;

  for ( i = 0; i < gNlang; i++ ) {
    j = 0;
    while ( gLang[i].ext[j++] ) {
      if ( !strcmp(ext,gLang[i].ext[j-1]) ) {
	this_lang = &gLang[i];
	break;
      }
      if ( this_lang ) break;
    }
  }

  return this_lang;
}



static void
print_langs ( void )
{
  int i;

  printf("Supported languages:\n");
  for ( i = 0; i < gNlang; i++ ) {
    printf("\t%-10s (%s)\n",gLang[i].lang,gLang[i].name);
  }
}
