#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <errno.h>
#include <string.h>
#include "syntax.h"

/* static helper functions */

static char *sh_check    ( sh_data *d );
static char *sh_grow     ( sh_data *d );
static char *sh_grow_by  ( sh_data *d, int );



void
sh_reset ( sh_data *d )
{
  /*  fprintf(stderr,"sh_reset\n"); */
  
  if ( d->buf ) free(d->buf);
  d->buf = NULL;
  d->len = 0;
  d->cap = 0;
}


/* 
   This function operates on a global buffer string.  It works with 
   a buffer of length BUFSIZ (usually 8 KB).  what is either CPP, 
   VERILOG, or VHDL.
*/

int
sh_cats ( sh_data *d, const char *fmt, ... )
{
  int     n;
  int     avail;
  char    buf[BUFSIZ];
  char    *ebuf = NULL;
  char    *nbuf;
  va_list ap;

  if ( !sh_check(d) ) return -1;

  va_start(ap, fmt);
  if ( (n = vsnprintf (buf,BUFSIZ,fmt,ap)) <= BUFSIZ ) {
    
    /* we had enough space */
    
    nbuf = &buf[0];
    
  } else if ( n > BUFSIZ ) {
    
    /* 
       not enough space in buf, need a dynamically allocated buffer
       (this should almost never happen) 
    */
    
    if ( (ebuf = malloc(n+1)) ) {
      n    = vsnprintf(ebuf,n,fmt,ap);
      nbuf = ebuf;
    } else {
      fprintf(stderr,"sh_cats: malloc: %s\n",strerror(errno));
      return -1;
    }
  }
  
  va_end(ap);
  
  /* 
     nbuf contains the data we need to concatenate onto d->buf,
     and n is the number of bytes it will occupy.
  */
  
  avail = d->cap - d->len - 1;  /* trailing '\0' */
  
  if ( n > avail && !sh_grow_by(d,n) ) {
    
    /* if we can't grow, bail out. */
    
    fprintf(stderr,"sh_cats: unable to allocate %d more bytes!\n",n);
    return -1;
  }
  
  /* now we definitely have enough space to copy */
  
  strcpy(&d->buf[d->len],nbuf);  /* strcpy includes '\0' */
  d->len += n;

  /* if we had to allocate memory, free it now. */
  
  if ( nbuf == ebuf ) free(ebuf);

  /* return the number of bytes we added. */
  
  return n;
}


int
sh_catc ( sh_data *d, int c )
{
  char *htmlent = NULL;

  /* Check where c needs to be a special html entity */
  switch ( c ) {
    case '<':
      htmlent = "&lt;";
      break;
    case '>':
      htmlent = "&gt;";
      break;
    case '&':
      htmlent = "&amp;";
      break;
    case '%':
      htmlent = "&#37;";
      break;
    case '\0':    /* don't bother adding NULL chars */
      return c;
  };

  if ( htmlent ) {
    sh_cats(d, htmlent);
    return c;
  }

  /* make sure we have space. */

  if ( !sh_check(d) ) return -1;

  /* try to grow if we need to */

  if ( d->len + 1 >= d->cap ) {
    if ( !sh_grow(d) ) return -1;
  }

  d->buf[d->len++] = c;
  d->buf[d->len]   = 0;

  return c;
}


/* 
   allocate a global buffer if necessary. 
*/

static char *
sh_check ( sh_data *d )
{
  if ( d->buf ) return d->buf;

  if ( (d->buf = calloc(1,BUFSIZ)) ) {
    d->cap = BUFSIZ;
    d->len = 0;
  }

  return d->buf;
}


/* 
   attempt to grow a global buffer by BUFSIZ bytes.
*/

static char *
sh_grow ( sh_data *d )
{
  return sh_grow_by(d,BUFSIZ);
}


static char *
sh_grow_by ( sh_data *d, int bytes )
{
  /* make sure we're called with valid parameters */

  if ( bytes <= 0 || 
       bytes >  (1<<20) )
    return NULL;

  /* 
     try to avoid calling this function every single time we need
     another byte. 
  */

  if ( bytes < BUFSIZ ) bytes = BUFSIZ;

  /*
    fprintf(stderr,"sh_grow_by: %p from %d to %d\n",d->buf,
            d->cap, d->cap+bytes);
  */

  if ( (d->buf = realloc(d->buf,d->cap+bytes)) ) {
    d->cap += bytes;
  } else { 
    fprintf(stderr,"sh_grow_by(%d): realloc: %s\n",
	    bytes, strerror(errno));
  }

  /*
    fprintf(stderr,"done sh_grow_by: %p\n",d->buf);
  */

  return d->buf;
}


